"""Main plotting tools."""
import numpy
import matplotlib.pyplot as plt
from matplotlib.gridspec import (GridSpec as GS,
                                 GridSpecFromSubplotSpec as sGS)
from anesthetic.gui.widgets import (Widget, Slider, Button,
                                    RadioButtons, TrianglePlot, CheckButtons)


class Higson(Widget):
    """Higson plot as shown in https://arxiv.org/abs/1703.09701 .

    Attributes
    ----------
        curve: matplotlib.lines.Line2D
            points currently plotted as a curve.

        point: matplotlib.lines.Line2D
            large indicator point currently plotted on the curve.

    """

    def __init__(self, fig, gridspec):
        super(Higson, self).__init__(fig, gridspec)
        self.ax.set_yticks([])
        self.ax.set_ylim(-0.1, 1.1)
        self.ax.set_ylabel(r'$LX$')
        self.ax.set_xlabel(r'$\log X$')

        self.curve, = self.ax.plot([None], [None], 'k-')
        self.point, = self.ax.plot([None], [None], 'ko')

    def update(self, logX, LX, i):
        """Update the line and the point in the higson plot.

        Parameters
        ----------
            logX: array-like
                log-volume compression values to plot

            LX: array-like
                Likelihood * volume compression

            i: int
                Current location of higson point

        """
        self.point.set_xdata(logX[i])
        self.point.set_ydata(LX[i])
        self.curve.set_xdata(logX)
        self.curve.set_ydata(LX)

    def reset_range(self):
        """Reset the ranges of the higson plot."""
        xdata = self.curve.get_xdata()
        self.ax.set_xlim(max(xdata), min(xdata))


class Evolution(Slider):
    """Slider controlling the evolution stage of the live points."""

    def __init__(self, fig, gridspec, action, valmax):
        super(Evolution, self).__init__(fig, gridspec, action, '',
                                        0, valmax, 0, 'horizontal')
        self.slider.valtext.set_horizontalalignment('right')
        self.slider.valtext.set_position((0.98, 0.5))

    def __call__(self):
        """Return the current iteration as an integer."""
        mx = self.slider.valmax-1
        val = int(super(Evolution, self).__call__())
        return min(val, mx)

    def set_text(self, logL, n):
        """Set the text at end of slider.

        Parameters
        ----------
            logL: float
                Current loglikelihood of evolution stage

            n: int
                Current number of live points of evolution stage

        """
        text = r'$\log L$: %.6g, $n_\mathrm{live}$: %i' % (logL, n)
        return super(Evolution, self).set_text(text)


class Temperature(Slider):
    """Logarithmic slider controlling temperature of the posterior points."""

    def __init__(self, fig, gridspec, action):
        super(Temperature, self).__init__(fig, gridspec, action, r'$kT$',
                                          -1, 5, 0, 'vertical')

    def __call__(self):
        """Return the current temperature."""
        return 10**super(Temperature, self).__call__()

    def set_text(self, kT):
        """Set the text at end of slider.

        Args:
            kT: float
                Current temperature of posterior points stage

        """
        text = r'%.2g' % kT
        return super(Temperature, self).set_text(text)


class RunPlotter(object):
    """Construct a control panel of information on a nested sampling run.

    Parameters
    ----------
        samples: anesthetic.samples.NestedSamples
            The root string for the chains files to be used, or a set of nested
            samples.

    Attributes
    ----------
        samples: anesthetic.samples.NestedSamples
            Object for extracting nested sampling data from chains files.

        fig: matplotlib.figure.Figure
            Reference to the underlying figure

        triangle: anesthetic.gui.widgets.TrianglePlot
            Corner plot of live or posterior samples.

        temperature: anesthetic.gui.plot.Temperature
            Slider selecting the posterior temperature.

        evolution: anesthetic.gui.plot.Evolution
            Slider selecting the live iteration.

        higson: anesthetic.gui.plot.Higson
            Higson plot of posterior weights.

        reset: anesthetic.gui.widgets.Button
            Button that resets the parameter ranges.

        reload: anesthetic.gui.widgets.Button
            Button that reloads the files.

        type: anesthetic.gui.widgets.RadioButtons
            Radio buttons that selects whether to plot live or posteriors.

        param_choice: anesthetic.gui.widgets.CheckButtons
            Checkbox that selects which parameters to plot.

    """

    def __init__(self, samples, params=None):
        self.samples = samples

        if params:
            self.params = numpy.array(params)
        else:
            self.params = numpy.array(self.samples.columns[:10])

        self.fig = plt.figure()
        self._set_up()
        self.redraw(None)

    def _set_up(self):
        """Draw the control panel.

        We implement the control panel using sequential recursive gridspecs.

        +----------------------------------------------+------+
        |                                              |      |
        |                  gs0[0]                      |gs0[1]|
        |                                              |      |
        |                                              |      |
        |                                              |      |
        |                                              |      |
        |                                              |      |
        |                                              |      |
        +-------------------------+-----------------+--+------+
        |                         |    gs11[0]      |         |
        |         gs10[0]         +-----------------+  gs1[2] |
        +-------------------------+    gs11[1]      |         |
        |                         +-----------------+         |
        |         gs10[1]         |    gs11[2]      |         |
        +-------------------------+-----------------+---------+

        These variable names are included in the __init__ function, and are
        named with an intuitive Huffman coding.

        """
        gs = GS(2, 1, height_ratios=[3, 1])
        gs0 = sGS(1, 2, width_ratios=[19, 1], subplot_spec=gs[0])
        gs1 = sGS(1, 3, width_ratios=[4, 1, 1], subplot_spec=gs[1])
        gs10 = sGS(2, 1, height_ratios=[1, 4], subplot_spec=gs1[0])
        gs11 = sGS(3, 1, height_ratios=[1, 1, 2], subplot_spec=gs1[1])

        self.triangle = TrianglePlot(self.fig, gs0[0])
        self.temperature = Temperature(self.fig, gs0[1], self.update)
        self.evolution = Evolution(self.fig, gs10[0],
                                   self.update, len(self.samples))
        self.higson = Higson(self.fig, gs10[1])
        self.reset = Button(self.fig, gs11[0],
                            self.reset_range, 'Reset Range')
        self.reload = Button(self.fig, gs11[1],
                             self.reload_file, 'Reload File')
        self.type = RadioButtons(self.fig, gs11[2],
                                 ('live', 'posterior'), self.update)
        self.param_choice = CheckButtons(self.fig, gs1[2],
                                         self.params, self.redraw)

    def redraw(self, _):
        """Redraw the triangle plot upon parameter updating."""
        self.triangle.draw(self.param_choice(), self.samples.tex)
        self.update(None)
        self.reset_range(None)
        self.fig.tight_layout()
        self.fig.canvas.draw()

    def points(self, label):
        """Get sample coordinates from nested sampling samples.

        Parameters
        ----------
            label: str
                label indicating the coordinate to extract.

        Returns
        -------
            array-like:
                sample 'label'-coordinates.

        """
        if self.type() == 'posterior':
            kT = self.temperature()
            return self.samples.posterior_points(1/kT)[label]
        else:
            i = self.evolution()
            logL = self.samples.logL.iloc[i]
            return self.samples.live_points(logL)[label]

    def update(self, _):
        """Update all the plots upon slider changes."""
        logX = numpy.log(self.samples.nlive/(self.samples.nlive+1)).cumsum()
        kT = self.temperature()
        LX = self.samples.logL/kT + logX
        LX = numpy.exp(LX-LX.max())
        i = self.evolution()
        logL = self.samples.logL.iloc[i]
        try:
            n = self.samples.nlive.iloc[i]
        except IndexError:
            n = 0

        self.triangle.update(self.points)

        self.evolution.set_text(logL, n)
        self.temperature.set_text(kT)

        self.higson.update(logX, LX, i)
        self.fig.canvas.draw()

    def reload_file(self, _):
        """Reload the data from file."""
        self.samples._reload_data()
        self.evolution.reset_range(valmax=len(self.samples))
        self.update(None)

    def reset_range(self, _):
        """Reset the parameter ranges."""
        self.triangle.reset_range()
        self.higson.reset_range()
        self.fig.canvas.draw()



"""Generic nested sampling reading tools."""
import os
from anesthetic.read.polychord import PolyChordReader
from anesthetic.read.multinest import MultiNestReader


def NestedReader(root):
    """Read from a variety of nested sampling chains."""
    mn = MultiNestReader(root)
    if os.path.isfile(mn.ev_file) and os.path.isfile(mn.phys_live_file):
        return mn

    pc = PolyChordReader(root)
    if os.path.isfile(pc.birth_file):
        return pc

    errstr = "Could not find nested sampling chains with root %s" % root
    try:
        raise FileNotFoundError(errstr)
    except NameError:
        raise IOError(errstr)



"""Widget wrappers to matplotlib.

These extend the matplotlib widgets by plotting themselves onto an axis and
storing a reference to both the widget object and the axis on which they are
plotted.

"""

from matplotlib.widgets import Button as mplButton
from matplotlib.widgets import CheckButtons as mplCheckButtons
from matplotlib.widgets import RadioButtons as mplRadioButtons
from anesthetic.gui._matplotlib import Slider as mplSlider
from anesthetic.utils import histogram
from anesthetic.plot import make_2d_axes


class Widget(object):
    """Base class for anesthetic gui widgets.

    Stores a reference to the underlying figure, the gridspec that the
    widget is placed at and the axes of the widget.

    Parameters
    ----------
        fig: matplotlib.figure.Figure
            Figure for drawing widget on.

        gridspec: matplotlib.gridspec.GridSpec
            Specification for where to draw in the figure.
            Technically could be any argument that can be passed to
            matplotlib.figure.Figure.add_subplot.

    Attributes
    ----------
        fig: (matplotlib.figure.Figure
            Figure for drawing widget on.

        gridspec: matplotlib.gridspec.GridSpec
            Specification for where to draw in the figure.
            Technically could be any argument that can be passed to
            matplotlib.figure.Figure.add_subplot.

        ax: matplotlib.axes.Axes
            Axes of widget.

    """

    def __init__(self, fig, gridspec):
        self.fig = fig
        self.gridspec = gridspec
        self.ax = self.fig.add_subplot(self.gridspec)


class LabelsWidget(Widget):
    """Widget with labels to choose from.

    Parameters
    ----------
        labels: list(str)
            Set of labels to be tied to Widget.

    Attributes
    ----------
        labels: list(str)
            Set of labels on Widget.

    """

    def __init__(self, fig, gridspec, labels):
        super(LabelsWidget, self).__init__(fig, gridspec)
        self.labels = labels


class Button(Widget):
    """Push button that performs an action.

    Parameters
    ----------
        action: func
            What should be run upon clicking the button.

        text: str
            Overlay text on the button.

    Attributes
    ----------
        button: matplotlib.widgets.Button
            matplotlib Button.

    """

    def __init__(self, fig, gridspec, action, text):
        super(Button, self).__init__(fig, gridspec)
        self.button = mplButton(self.ax, text)
        self.button.on_clicked(action)


class CheckButtons(LabelsWidget):
    """Set of checkboxes.

    Defaults to choosing the only the first label at start.

    Parameters
    ----------
        action: func
            What should be done upon checking the box.

    Attributes
    ----------
        buttons: matplotlib.widgets.CheckButtons
            matplotlib CheckButtons.

    """

    def __init__(self, fig, gridspec, labels, action):
        super(CheckButtons, self).__init__(fig, gridspec, labels)
        default = [i == 0 for i, _ in enumerate(self.labels)]
        self.buttons = mplCheckButtons(self.ax, self.labels, default)
        self.buttons.on_clicked(action)

    def __call__(self):
        """Get current status of the check boxes."""
        return self.labels[list(self.buttons.get_status())]


class RadioButtons(LabelsWidget):
    """Set of radio selection choices.

    Parameters
    ----------
        labels: list(str))

    Attributes
    ----------
        buttons: matplotlib.widgets.RadioButtons
            matplotlib RadioButtons.

    """

    def __init__(self, fig, gridspec, labels, action):
        super(RadioButtons, self).__init__(fig, gridspec, labels)
        self.buttons = mplRadioButtons(self.ax, self.labels)
        self.buttons.on_clicked(action)

    def __call__(self):
        """Get current selection string."""
        return self.buttons.value_selected


class Slider(Widget):
    """Choose a parameter via a slider widget.

    Parameters
    ----------
        action: func
            What should be done upon altering the slider.

        text: str
            Text at the start of the slider.

        valmin, valmax, valinit: float
            Range and initial value for the slider.

        orientation: str
            Orientation for slider ('horizontal' or 'vertical').

    Attributes
    ----------
        slider: matplotlib.widgets.Slider
            matplotlib Slider.

    """

    def __init__(self, fig, gridspec, action, text,
                 valmin, valmax, valinit, orientation):
        super(Slider, self).__init__(fig, gridspec)
        self.slider = mplSlider(self.ax, text, valmin, valmax, valinit=valinit,
                                orientation=orientation)
        self.slider.on_changed(action)

    def set_text(self, text):
        """Set the text at the end of the slider.

        Parameters
        ----------
            text: str
                Text to be chosen

        """
        self.slider.valtext.set_text(text)

    def reset_range(self, valmin=None, valmax=None):
        """Reset the range of the slider.

        Kwargs:
            valmin, valmax: (float), optional:
                The new limits to the slider.

        """
        if valmax is not None:
            self.slider.valmax = valmax
        if valmin is not None:
            self.slider.valmin = valmin
        self.slider.ax.set_xlim(self.slider.valmin, self.slider.valmax)

    def __call__(self):
        """Return the float value in the slider."""
        return self.slider.val


class TrianglePlot(Widget):
    """Triangle plot widget as exemplified by getdist and corner.py.

    For other examples of these plots, see:
    https://getdist.readthedocs.io
    https://corner.readthedocs.io

    Attributes
    ----------
        ax: pandas.DataFrame(matplotlib.axes.Axes)
            Mapping from pairs of parameters to axes for plotting.
    """

    def __init__(self, fig, gridspec):
        super(TrianglePlot, self).__init__(fig, gridspec)
        self.fig.delaxes(self.ax)
        _, self.ax = make_2d_axes([], fig=self.fig, subplot_spec=self.gridspec)

    def draw(self, labels, tex={}):
        """Draw a new triangular grid for list of parameters labels.

        Parameters
        ----------
            labels: list(str)
                labels for the triangular grid.

        """
        # Remove any existing axes
        for y, row in self.ax.iterrows():
            for x, ax in row.iteritems():
                if ax is not None:
                    if x == y:
                        self.fig.delaxes(ax.twin)
                    self.fig.delaxes(ax)

        # Set up the axes
        _, self.ax = make_2d_axes(labels, upper=False, tex=tex,
                                  fig=self.fig, subplot_spec=self.gridspec)

        # Plot no points  points.
        for y, row in self.ax.iterrows():
            for x, ax in row.iteritems():
                if ax is not None:
                    if x == y:
                        ax.twin.plot([None], [None], 'k-')
                    else:
                        ax.plot([None], [None], 'k.')

    def update(self, f):
        """Update the points in the triangle plot using f function.

        Parameters
        ----------
            f: func: str -> array(float)
                this function should take in a parameter label i, and return an
                array-like object of the i-coordinate of the samples

        """
        for y, row in self.ax.iterrows():
            for x, ax in row.iteritems():
                if ax is not None:
                    if x == y:
                        datx, daty = histogram(f(x), bins='auto')
                        ax.twin.lines[0].set_xdata(datx)
                        ax.twin.lines[0].set_ydata(daty)
                    else:
                        ax.lines[0].set_xdata(f(x))
                        ax.lines[0].set_ydata(f(y))

    def reset_range(self):
        """Reset the range of each grid."""
        for y, row in self.ax.iterrows():
            for x, ax in row.iteritems():
                if ax is not None:
                    if x == y:
                        ax.twin.relim()
                        ax.twin.autoscale_view()
                    ax.relim()
                    ax.autoscale_view()



"""Tools for reading from getdist chains files."""
import numpy
import glob
from anesthetic.read.base import ChainReader


class GetDistReader(ChainReader):
    """Read getdist files."""

    def paramnames(self):
        r"""Read <root>.paramnames in getdist format.

        This file should contain one or two columns. The first column indicates
        a reference name for the sample, used as labels in the pandas array.
        The second optional column should include the equivalent axis label,
        possibly in tex, with the understanding that it will be surrounded by
        dollar signs, for example

        <root.paramnames>

        a1     a_1
        a2     a_2
        omega  \omega
        """
        try:
            with open(self.paramnames_file, 'r') as f:
                paramnames = []
                tex = {}
                for line in f:
                    line = line.strip().split()
                    paramname = line[0].replace('*', '')
                    paramnames.append(paramname)
                    if len(line) > 1:
                        tex[paramname] = '$' + ' '.join(line[1:]) + '$'
                return paramnames, tex
        except IOError:
            return super(GetDistReader, self).paramnames()

    def limits(self):
        """Read <root>.ranges in getdist format."""
        try:
            with open(self.ranges_file, 'r') as f:
                limits = {}
                for line in f:
                    line = line.strip().split()
                    paramname = line[0]
                    try:
                        xmin = float(line[1])
                    except ValueError:
                        xmin = None
                    try:
                        xmax = float(line[2])
                    except ValueError:
                        xmax = None
                    limits[paramname] = (xmin, xmax)
                return limits
        except IOError:
            return super(GetDistReader, self).limits()

    def samples(self):
        """Read <root>_1.txt in getdist format."""
        data = numpy.concatenate([numpy.loadtxt(chains_file)
                                  for chains_file in self.chains_files])
        weights, chi2, samples = numpy.split(data, [1, 2], axis=1)
        logL = chi2/-2.
        return weights.flatten(), logL.flatten(), samples

    @property
    def paramnames_file(self):
        """File containing parameter names."""
        return self.root + '.paramnames'

    @property
    def ranges_file(self):
        """File containing parameter names."""
        return self.root + '.ranges'

    @property
    def chains_files(self):
        """File containing parameter names."""
        files = glob.glob(self.root + '_[0-9].txt')
        if not files:
            files = [self.root + '.txt']

        return files



"""Main classes for the anesthetic module.

- ``MCMCSamples``
- ``NestedSamples``
"""
import numpy
import pandas
from scipy.special import logsumexp
from anesthetic.plot import (make_1d_axes, make_2d_axes, plot_1d,
                             scatter_plot_2d, contour_plot_2d)
from anesthetic.read.getdist import GetDistReader
from anesthetic.read.nested import NestedReader
from anesthetic.utils import compute_nlive
from anesthetic.gui.plot import RunPlotter
from anesthetic.weighted_pandas import WeightedDataFrame


class MCMCSamples(WeightedDataFrame):
    """Storage and plotting tools for MCMC samples.

    Extends the pandas.DataFrame by providing plotting methods and
    standardising sample storage.

    Example plotting commands include

    - ``mcmc.plot_1d(['paramA', 'paramB'])``
    - ``mcmc.plot_2d(['paramA', 'paramB'])``
    - ``mcmc.plot_2d(['paramA', 'paramB'], ['paramC', 'paramD'])``

    Parameters
    ----------
    root: str, optional
        root for reading chains from file. Overrides all other arguments.

    data: numpy.array
        Coordinates of samples. shape = (nsamples, ndims).

    columns: list(str)
        reference names of parameters

    w: numpy.array
        weights of samples.

    logL: numpy.array
        loglikelihoods of samples.

    tex: dict
        mapping from coloumns to tex labels for plotting

    limits: dict
        mapping from coloumns to prior limits

    """

    _metadata = WeightedDataFrame._metadata + ['tex', 'limits', 'u', 'root']

    @property
    def _constructor(self):
        return MCMCSamples

    def __init__(self, *args, **kwargs):
        self.root = kwargs.pop('root', None)
        if self.root is not None:
            reader = GetDistReader(self.root)
            w, logL, samples = reader.samples()
            params, tex = reader.paramnames()
            limits = reader.limits()
            self.__init__(data=samples, columns=params, w=w, logL=logL,
                          tex=tex, limits=limits)
        else:
            logL = kwargs.pop('logL', None)
            self.tex = kwargs.pop('tex', {})
            self.limits = kwargs.pop('limits', {})
            self.u = None

            super(MCMCSamples, self).__init__(*args, **kwargs)

            self.u = numpy.random.rand(len(self))

            if self.w is not None:
                self['weight'] = self.w
                self.tex['weight'] = r'MCMC weight'

            if logL is not None:
                self['logL'] = logL
                self.tex['logL'] = r'$\log\mathcal{L}$'

    def plot(self, ax, paramname_x, paramname_y=None, *args, **kwargs):
        """Interface for 2D and 1D plotting routines.

        Produces a single 1D or 2D plot on an axis.

        Parameters
        ----------
        ax: matplotlib.axes.Axes
            Axes to plot on

        paramname_x: str
            Choice of parameter to plot on x-coordinate from self.columns.

        paramname_y: str
            Choice of parameter to plot on y-coordinate from self.columns.
            If not provided, or the same as paramname_x, then 1D plot produced.

        plot_type: str, optional
            Must be in {'kde','scatter'}. (Default: 'kde')

        Returns
        -------
        fig: matplotlib.figure.Figure
            New or original (if supplied) figure object

        axes: pandas.DataFrame or pandas.Series of matplotlib.axes.Axes
            Pandas array of axes objects

        """
        plot_type = kwargs.pop('plot_type', 'kde')

        if paramname_y is None or paramname_x == paramname_y:
            xmin, xmax = self._limits(paramname_x)
            return plot_1d(ax, self[paramname_x].compress(),
                           xmin=xmin, xmax=xmax,
                           *args, **kwargs)

        xmin, xmax = self._limits(paramname_x)
        ymin, ymax = self._limits(paramname_y)

        if plot_type == 'kde':
            nsamples = None
            plot = contour_plot_2d
        elif plot_type == 'scatter':
            nsamples = 500
            plot = scatter_plot_2d
        else:
            raise ValueError("plot_type must be in {'kde', 'scatter'}")

        return plot(ax, self[paramname_x].compress(nsamples),
                    self[paramname_y].compress(nsamples),
                    xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax,
                    *args, **kwargs)

    def plot_1d(self, axes, *args, **kwargs):
        """Create an array of 1D plots.

        Parameters
        ----------
        axes: plotting axes
            Can be either:
                * list(str) or str
                * pandas.Series(matplotlib.axes.Axes)
            If a pandas.Series is provided as an existing set of axes, then
            this is used for creating the plot. Otherwise a new set of axes are
            created using the list or lists of strings.

        Returns
        -------
        fig: matplotlib.figure.Figure
            New or original (if supplied) figure object

        axes: pandas.Series of matplotlib.axes.Axes
            Pandas array of axes objects

        """
        if not isinstance(axes, pandas.Series):
            fig, axes = make_1d_axes(axes, tex=self.tex)
        else:
            fig = axes.values[~axes.isna()][0].figure

        for x, ax in axes.iteritems():
            if ax is not None and x in self:
                self.plot(ax, x, *args, **kwargs)

        return fig, axes

    def plot_2d(self, axes, *args, **kwargs):
        """Create an array of 2D plots.

        To avoid intefering with y-axis sharing, one-dimensional plots are
        created on a separate axis, which is monkey-patched onto the argument
        ax as the attribute ax.twin.

        Parameters
        ----------
        axes: plotting axes
            Can be either:
                * list(str) if the x and y axes are the same
                * [list(str),list(str)] if the x and y axes are different
                * pandas.DataFrame(matplotlib.axes.Axes)
            If a pandas.DataFrame is provided as an existing set of axes, then
            this is used for creating the plot. Otherwise a new set of axes are
            created using the list or lists of strings.

        types: list(str) or str, optional
            What type (or types) of plots to produce. If two types are provided
            then pairs of parameters 'above the diagonal' have the second type.
            each string must be one of {'kde', 'scatter'}.
            Default: ['kde', 'scatter']

        Returns
        -------
        fig: matplotlib.figure.Figure
            New or original (if supplied) figure object

        axes: pandas.DataFrame of matplotlib.axes.Axes
            Pandas array of axes objects

        """
        types = kwargs.pop('types', ['kde', 'scatter'])

        if not isinstance(axes, pandas.DataFrame):
            upper = None if len(types) > 1 else False
            fig, axes = make_2d_axes(axes, tex=self.tex, upper=upper)
        else:
            fig = axes.values[~axes.isna()][0].figure

        types = numpy.atleast_1d(types)

        for y, row in axes.iterrows():
            for x, ax in row.iteritems():
                if ax is not None and x in self and y in self:
                    plot_type = types[-1] if ax._upper else types[0]
                    ax_ = ax.twin if x == y else ax
                    self.plot(ax_, x, y, plot_type=plot_type, *args, **kwargs)

        return fig, axes

    def _limits(self, paramname):
        return self.limits.get(paramname, (None, None))

    def _reload_data(self):
        self.__init__(root=self.root)
        return self


class NestedSamples(MCMCSamples):
    """Storage and plotting tools for Nested Sampling samples.

    We extend the MCMCSamples class with the additional methods:

    * ``self.ns_output()``
    * ``self.live_points(logL)``
    * ``self.posterior_points(beta)``

    Parameters
    ----------
    root: str, optional
        root for reading chains from file. Overrides all other arguments.

    data: numpy.array
        Coordinates of samples. shape = (nsamples, ndims).

    columns: list(str)
        reference names of parameters

    logL: numpy.array
        loglikelihoods of samples.

    logL_birth: numpy.array or int
        birth loglikelihoods, or number of live points.

    tex: dict
        mapping from coloumns to tex labels for plotting

    limits: dict
        mapping from coloumns to prior limits

    beta: float
        thermodynamic temperature

    """

    _metadata = MCMCSamples._metadata + ['_beta']

    @property
    def _constructor(self):
        return NestedSamples

    def __init__(self, *args, **kwargs):
        self.root = kwargs.pop('root', None)
        self._beta = kwargs.pop('beta', 1.)
        if self.root is not None:
            reader = NestedReader(self.root)
            samples, logL, logL_birth = reader.samples()
            params, tex = reader.paramnames()
            limits = reader.limits()
            self.__init__(data=samples, columns=params,
                          logL=logL, logL_birth=logL_birth,
                          tex=tex, limits=limits)
        else:
            logL_birth = kwargs.pop('logL_birth', None)
            super(NestedSamples, self).__init__(*args, **kwargs)

            # Compute nlive
            if logL_birth is not None:
                if isinstance(logL_birth, int):
                    nlive = logL_birth
                    self['nlive'] = nlive
                    descending = numpy.arange(nlive, 0, -1)
                    self.loc[len(self)-nlive:, 'nlive'] = descending
                else:
                    self['logL_birth'] = logL_birth
                    self.tex['logL_birth'] = r'$\log\mathcal{L}_{\rm birth}$'
                    self['nlive'] = compute_nlive(self.logL, self.logL_birth)

                self.tex['nlive'] = r'$n_{\rm live}$'

            if 'nlive' in self:
                self.beta = self._beta

            if self.w is not None:
                self['weight'] = self.w
                self.tex['weight'] = r'MCMC weight'

    @property
    def beta(self):
        """Thermodynamic inverse temperature."""
        return self._beta

    @beta.setter
    def beta(self, beta):
        self._beta = beta
        logw = self._dlogX() + self.beta*self.logL
        self.w = numpy.exp(logw - logw.max())

    def set_beta(self, beta, inplace=False):
        """Change the inverse temperature.

        Parameters
        ----------
        beta: float
            Temperature to set

        inplace: bool, optional
            Indicates whether to modify the existing array, or return a copy
            with the temperature changed. Default: False

        """
        if inplace:
            self.beta = beta
        else:
            data = self.copy()
            data.beta = beta
            return data

    def ns_output(self, nsamples=200):
        """Compute Bayesian global quantities.

        Using nested sampling we can compute the evidence (logZ),
        Kullback-Leibler divergence (D) and Bayesian model dimensionality (d).
        More precisely, we can infer these quantities via their probability
        distribution.

        Parameters
        ----------
        nsamples: int, optional
            number of samples to generate (Default: 100)

        Returns
        -------
        pandas.DataFrame
            Samples from the P(logZ, D, d) distribution

        """
        dlogX = self._dlogX(nsamples)

        logZ = logsumexp(self.logL.values + dlogX, axis=1)
        logw = ((self.logL.values + dlogX).T - logZ).T
        S = ((self.logL.values + numpy.zeros_like(dlogX)).T
             - logZ).T

        D = numpy.exp(logsumexp(logw, b=S, axis=1))
        d = numpy.exp(logsumexp(logw, b=(S.T-D).T**2, axis=1))*2

        samples = numpy.vstack((logZ, D, d)).T
        params = ['logZ', 'D', 'd']
        tex = {'logZ': r'$\log\mathcal{Z}$',
               'D': r'$\mathcal{D}$',
               'd': r'$d$'}
        return MCMCSamples(data=samples, columns=params, tex=tex)

    def live_points(self, logL):
        """Get the live points within logL."""
        return self[(self.logL > logL) & (self.logL_birth <= logL)]

    def posterior_points(self, beta):
        """Get the posterior points at temperature beta."""
        return self.set_beta(beta).compress(0)

    def gui(self, params=None):
        """Construct a graphical user interface for viewing samples."""
        return RunPlotter(self, params)

    def _dlogX(self, nsamples=None):
        """Compute volume of shell of loglikelihood.

        Parameters
        ----------
        nsamples: int, optional
            Number of samples to generate. optional. If None, then compute the
            statistical average. If integer, generate samples from the
            distribution. (Default: None)

        """
        if nsamples is None:
            t = numpy.atleast_2d(numpy.log(self.nlive/(self.nlive+1)))
            nsamples = 1
        else:
            t = numpy.log(numpy.random.rand(nsamples, len(self))
                          )/self.nlive.values
        logX = numpy.concatenate((numpy.zeros((nsamples, 1)),
                                  t.cumsum(axis=1),
                                  -numpy.inf*numpy.ones((nsamples, 1))
                                  ), axis=1)
        dlogX = logsumexp([logX[:, :-2], logX[:, 2:]],
                          b=[numpy.ones_like(t), -numpy.ones_like(t)], axis=0)
        dlogX -= numpy.log(2)
        return numpy.squeeze(dlogX)



"""Anesthetic: nested sampling post-processing.

Key routines:

- ``MCMCSamples.build``
- ``MCMCSamples.read``
- ``NestedSamples.build``
- ``NestedSamples.read``

"""
import anesthetic.samples
import anesthetic.plot


MCMCSamples = anesthetic.samples.MCMCSamples
NestedSamples = anesthetic.samples.NestedSamples
make_2d_axes = anesthetic.plot.make_2d_axes
make_1d_axes = anesthetic.plot.make_1d_axes
get_legend_proxy = anesthetic.plot.get_legend_proxy



"""Tools for reading from polychord chains files."""
import numpy
from anesthetic.read.getdist import GetDistReader


class PolyChordReader(GetDistReader):
    """Read polychord files."""

    def samples(self):
        """Read ``<root>_dead-birth.txt`` in polychord format."""
        data = numpy.loadtxt(self.birth_file)
        samples, logL, logL_birth = numpy.split(data, [-2, -1], axis=1)
        return samples, logL.flatten(), logL_birth.flatten()

    @property
    def birth_file(self):
        """File containing dead and birth contours."""
        return self.root + '_dead-birth.txt'



"""Data-processing utility functions."""
import numpy
import pandas
from scipy.interpolate import interp1d


def channel_capacity(w):
    r"""Channel capacity (effective sample size).

    .. math::

        H = \sum_i p_i \log p_i

        p_i = \frac{w_i}{\sum_j w_j}

        N = e^{-H}
    """
    with numpy.errstate(divide='ignore', invalid='ignore'):
        W = numpy.array(w)/sum(w)
        H = numpy.nansum(numpy.log(W)*W)
        return numpy.exp(-H)


def compress_weights(w, u=None, nsamples=None):
    """Compresses weights to their approximate channel capacity."""
    if u is None:
        u = numpy.random.rand(len(w))

    if w is None:
        w = numpy.ones_like(u)

    if nsamples is not None:
        W = w/w.max()
        if nsamples > 0 and sum(W) > nsamples:
            W = W/sum(W)*nsamples
    else:
        W = w * channel_capacity(w) / w.sum()

    fraction, integer = numpy.modf(W)
    extra = (u < fraction).astype(int)
    return (integer + extra).astype(int)


def quantile(a, q, w):
    """Compute the weighted quantile for a one dimensional array."""
    if w is None:
        w = numpy.ones_like(a)
    i = numpy.argsort(a)
    c = numpy.cumsum(w[i])
    c /= c[-1]
    icdf = interp1d(c/c[-1], a[i])
    return icdf(q)


def check_bounds(d, xmin=None, xmax=None):
    """Check if we need to apply strict bounds."""
    if xmin is not None and (d.min() - xmin)/(d.max()-d.min()) > 1e-2:
        xmin = None
    if xmax is not None and (xmax - d.max())/(d.max()-d.min()) > 1e-2:
        xmax = None
    return xmin, xmax


def mirror_1d(d, xmin=None, xmax=None):
    """If necessary apply reflecting boundary conditions."""
    if xmin is not None and xmax is not None:
        xmed = (xmin+xmax)/2
        return numpy.concatenate((2*xmin-d[d < xmed], d, 2*xmax-d[d >= xmed]))
    elif xmin is not None:
        return numpy.concatenate((2*xmin-d, d))
    elif xmax is not None:
        return numpy.concatenate((d, 2*xmax-d))
    else:
        return d


def mirror_2d(d_x_, d_y_, xmin=None, xmax=None, ymin=None, ymax=None):
    """If necessary apply reflecting boundary conditions."""
    d_x = d_x_.copy()
    d_y = d_y_.copy()

    if xmin is not None and xmax is not None:
        xmed = (xmin+xmax)/2
        d_y = numpy.concatenate((d_y[d_x < xmed], d_y, d_y[d_x >= xmed]))
        d_x = numpy.concatenate((2*xmin-d_x[d_x < xmed], d_x,
                                 2*xmax-d_x[d_x >= xmed]))
    elif xmin is not None:
        d_y = numpy.concatenate((d_y, d_y))
        d_x = numpy.concatenate((2*xmin-d_x, d_x))
    elif xmax is not None:
        d_y = numpy.concatenate((d_y, d_y))
        d_x = numpy.concatenate((d_x, 2*xmax-d_x))

    if ymin is not None and ymax is not None:
        ymed = (ymin+ymax)/2
        d_x = numpy.concatenate((d_x[d_y < ymed], d_x, d_x[d_y >= ymed]))
        d_y = numpy.concatenate((2*ymin-d_y[d_y < ymed], d_y,
                                 2*ymax-d_y[d_y >= ymed]))
    elif ymin is not None:
        d_x = numpy.concatenate((d_x, d_x))
        d_y = numpy.concatenate((2*ymin-d_y, d_y))
    elif ymax is not None:
        d_x = numpy.concatenate((d_x, d_x))
        d_y = numpy.concatenate((d_y, 2*ymax-d_y))

    return d_x, d_y


def nest_level(lst):
    """Calculate the nesting level of a list."""
    if not isinstance(lst, list):
        return 0
    if not lst:
        return 1
    return max(nest_level(item) for item in lst) + 1


def histogram(a, **kwargs):
    """Produce a histogram for path-based plotting.

    This is a cheap histogram. Necessary if one wants to update the histogram
    dynamically, and redrawing and filling is very expensive.

    This has the same arguments and keywords as numpy.histogram, but is
    normalised to 1.
    """
    hist, bin_edges = numpy.histogram(a, **kwargs)
    xpath, ypath = numpy.empty((2, 4*len(hist)))
    ypath[0::4] = ypath[3::4] = 0
    ypath[1::4] = ypath[2::4] = hist
    xpath[0::4] = xpath[1::4] = bin_edges[:-1]
    xpath[2::4] = xpath[3::4] = bin_edges[1:]
    mx = max(ypath)
    if mx:
        ypath /= max(ypath)
    return xpath, ypath


def compute_nlive(death, birth):
    """Compute number of live points from birth and death contours."""
    contours = numpy.concatenate(([birth[0]], death))
    index = numpy.arange(death.size)
    birth_index = contours.searchsorted(birth)-1
    births = pandas.Series(+1, index=birth_index).sort_index()
    deaths = pandas.Series(-1, index=index)
    nlive = pandas.concat([births, deaths]).sort_index()
    nlive = nlive.groupby(nlive.index).sum().cumsum()
    return nlive.values[:-1]


def unique(a):
    """Find unique elements, retaining order."""
    b = []
    for x in a:
        if x not in b:
            b.append(x)
    return b



"""Pandas DataFrame and Series with weighted samples."""

import numpy
import pandas
from anesthetic.utils import (compress_weights, channel_capacity, quantile)


class WeightedSeries(pandas.Series):
    """Weighted version of pandas.Series."""

    _metadata = ['w', 'u']

    @property
    def _constructor(self):
        return WeightedSeries

    def __init__(self, *args, **kwargs):
        self.w = kwargs.pop('w', None)
        self.u = kwargs.pop('u', None)
        super(WeightedSeries, self).__init__(*args, **kwargs)

    def mean(self):
        """Weighted mean of the sampled distribution."""
        return numpy.average(self, weights=self.w)

    def std(self):
        """Weighted standard deviation of the sampled distribution."""
        return numpy.sqrt(self.var())

    def var(self):
        """Weighted variance of the sampled distribution."""
        return numpy.average((self-self.mean())**2, weights=self.w)

    def median(self):
        """Weighted median of the sampled distribution."""
        return float(self.quantile())

    def quantile(self, q=0.5):
        """Weighted quantile of the sampled distribution."""
        return quantile(self.values, q, self.w)

    def neff(self):
        """Effective number of samples."""
        if self.w is None:
            return len(self)
        else:
            return channel_capacity(self.w)

    def compress(self, nsamples=None):
        """Reduce the number of samples by discarding low-weights.

        Parameter
        ---------
        neff: int, optional
            effective number of samples after compression. If not supplied,
            then reduce to the channel capacity (theoretical optimum
            compression). If <=0, then compress so that all weights are unity.

        """
        i = compress_weights(self.w, self.u, nsamples)
        return self.repeat(i)


class WeightedDataFrame(pandas.DataFrame):
    """Weighted version of pandas.DataFrame."""

    _metadata = ['w', 'u']

    @property
    def _constructor_sliced(self):
        def __constructor_sliced(*args, **kwargs):
            return WeightedSeries(*args, w=self.w, u=self.u, **kwargs)
        return __constructor_sliced

    @property
    def _constructor(self):
        return WeightedDataFrame

    def __init__(self, *args, **kwargs):
        self.w = kwargs.pop('w', None)
        self.u = kwargs.pop('u', None)
        super(WeightedDataFrame, self).__init__(*args, **kwargs)

    def mean(self):
        """Weighted mean of the sampled distribution."""
        return pandas.Series(numpy.average(self, weights=self.w, axis=0),
                             index=self.columns)

    def std(self):
        """Weighted standard deviation of the sampled distribution."""
        return numpy.sqrt(self.var())

    def var(self):
        """Weighted variance of the sampled distribution."""
        return pandas.Series(numpy.average((self-self.mean())**2,
                                           weights=self.w, axis=0),
                             index=self.columns)

    def cov(self):
        """Weighted covariance of the sampled distribution."""
        return pandas.DataFrame(numpy.cov(self.T, aweights=self.w),
                                index=self.columns, columns=self.columns)

    def median(self):
        """Weighted median of the sampled distribution."""
        return self.quantile()

    def quantile(self, q=0.5):
        """Weighted quantile of the sampled distribution."""
        data = numpy.array([c.quantile(q) for _, c in self.iteritems()])
        if numpy.isscalar(q):
            return pandas.Series(data, index=self.columns)
        else:
            return pandas.DataFrame(data.T, columns=self.columns, index=q)

    def neff(self):
        """Effective number of samples."""
        if self.w is None:
            return len(self)
        else:
            return channel_capacity(self.w)

    def compress(self, nsamples=None):
        """Reduce the number of samples by discarding low-weights.

        Parameter
        ---------
        neff: int, optional
            effective number of samples after compression. If not supplied,
            then reduce to the channel capacity (theoretical optimum
            compression). If <=0, then compress so that all weights are unity.

        """
        i = compress_weights(self.w, self.u, nsamples)
        data = numpy.repeat(self.values, i, axis=0)
        index = numpy.repeat(self.index.values, i)
        return pandas.DataFrame(data=data, index=index, columns=self.columns)



"""Lower-level plotting tools.

Routines that may be of use to users wishing for more fine-grained control may
wish to use.

- ``make_1d_axes``
- ``make_2d_axes``
- ``get_legend_proxy``

to create a set of axes and legend proxies.

"""
import numpy
import pandas
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec as GS, GridSpecFromSubplotSpec as SGS
from anesthetic.kde import kde_1d, kde_2d
from anesthetic.utils import check_bounds, nest_level, unique
from scipy.interpolate import interp1d
from matplotlib.ticker import MaxNLocator
from matplotlib.colors import LinearSegmentedColormap
from matplotlib.collections import PathCollection


def make_1d_axes(params, **kwargs):
    """Create a set of axes for plotting 1D marginalised posteriors.

    Parameters
    ----------
        params: list(str)
            names of parameters.

        tex: dict(str:str), optional
            Dictionary mapping params to tex plot labels.

        fig: matplotlib.figure.Figure, optional
            Figure to plot on.
            Default: matplotlib.pyplot.figure()

        ncols: int
            Number of columns in the plot
            option, default ceil(sqrt(num_params))

        subplot_spec: matplotlib.gridspec.GridSpec, optional
            gridspec to plot array as part of a subfigure
            Default: None

    Returns
    -------
    fig: matplotlib.figure.Figure
        New or original (if supplied) figure object

    axes: pandas.Series(matplotlib.axes.Axes)
        Pandas array of axes objects

    """
    axes = pandas.Series(index=numpy.atleast_1d(params), dtype=object)
    axes[:] = None
    tex = kwargs.pop('tex', {})
    fig = kwargs.pop('fig') if 'fig' in kwargs else plt.figure()
    ncols = kwargs.pop('ncols', int(numpy.ceil(numpy.sqrt(len(axes)))))
    nrows = int(numpy.ceil(len(axes)/float(ncols)))
    if 'subplot_spec' in kwargs:
        grid = SGS(nrows, ncols, wspace=0,
                   subplot_spec=kwargs.pop('subplot_spec'))
    else:
        grid = GS(nrows, ncols, wspace=0)

    if kwargs:
        raise TypeError('Unexpected **kwargs: %r' % kwargs)

    tex = {p: tex[p] if p in tex else p for p in axes.index}

    for p, g in zip(axes.index, grid):
        axes[p] = ax = fig.add_subplot(g)
        ax.set_xlabel(tex[p])
        ax.set_yticks([])

    for x, ax in axes.dropna().iteritems():
        ax.xaxis.set_major_locator(MaxNLocator(2, integer=True))

    return fig, axes


def make_2d_axes(params, **kwargs):
    """Create a set of axes for plotting 2D marginalised posteriors.

    Parameters
    ----------
        params: lists of parameters
            Can be either:
            * list(str) if the x and y axes are the same
            * [list(str),list(str)] if the x and y axes are different
            Strings indicate the names of the parameters

        tex: dict(str:str), optional
            Dictionary mapping params to tex plot labels.
            Default: params

        upper: None or logical, optional
            Whether to create plots in the upper triangle.
            If None do both. Default: None

        diagonal: True, optional
            Whether to create 1D marginalised plots on the diagonal.
            Default: True

        fig: matplotlib.figure.Figure, optional
            Figure to plot on.
            Default: matplotlib.pyplot.figure()

        subplot_spec: matplotlib.gridspec.GridSpec, optional
            gridspec to plot array as part of a subfigure.
            Default: None

    Returns
    -------
    fig: matplotlib.figure.Figure
        New or original (if supplied) figure object

    axes: pandas.DataFrame(matplotlib.axes.Axes)
        Pandas array of axes objects

    """
    if nest_level(params) == 2:
        xparams, yparams = params
    else:
        xparams = yparams = params
    axes = pandas.DataFrame(index=numpy.atleast_1d(yparams),
                            columns=numpy.atleast_1d(xparams),
                            dtype=object)
    axes[:][:] = None

    tex = kwargs.pop('tex', {})
    fig = kwargs.pop('fig') if 'fig' in kwargs else plt.figure()
    if 'subplot_spec' in kwargs:
        grid = SGS(*axes.shape, hspace=0, wspace=0,
                   subplot_spec=kwargs.pop('subplot_spec'))
    else:
        grid = GS(*axes.shape, hspace=0, wspace=0)

    upper = kwargs.pop('upper', None)
    diagonal = kwargs.pop('diagonal', True)

    if kwargs:
        raise TypeError('Unexpected **kwargs: %r' % kwargs)

    if axes.size == 0:
        return fig, axes

    tex = {p: tex[p] if p in tex else p
           for p in numpy.concatenate((axes.index, axes.columns))}

    all_params = list(axes.index) + list(axes.columns)

    for j, y in enumerate(axes.index):
        for i, x in enumerate(axes.columns):
            lower = not (x in axes.index and y in axes.columns
                         and all_params.index(x) > all_params.index(y))

            if x == y and not diagonal:
                continue

            if upper == lower and x != y:
                continue

            if axes[x][y] is None:
                sx = list(axes[x].dropna())
                sx = sx[0] if sx else None
                sy = list(axes.T[y].dropna())
                sy = sy[0] if sy else None
                axes[x][y] = fig.add_subplot(grid[j, i],
                                             sharex=sx, sharey=sy)
            if x == y:
                axes[x][y].twin = axes[x][y].twinx()
                axes[x][y].twin.set_yticks([])
                axes[x][y].twin.set_ylim(0, 1.1)

            axes[x][y]._upper = not lower

    for y, ax in axes.bfill(axis=1).iloc[:, 0].dropna().iteritems():
        ax.set_ylabel(tex[y])

    for x, ax in axes.ffill(axis=0).iloc[-1, :].dropna().iteritems():
        ax.set_xlabel(tex[x])

    for y, ax in axes.iterrows():
        ax_ = ax.dropna()
        if len(ax_):
            for a in ax_[1:]:
                a.tick_params('y', left=False, labelleft=False)

    for x, ax in axes.iteritems():
        ax_ = ax.dropna()
        if len(ax_):
            for a in ax_[:-1]:
                a.tick_params('x', bottom=False, labelbottom=False)

    for y, ax in axes.bfill(axis=1).iloc[:, 0].dropna().iteritems():
        ax.yaxis.set_major_locator(MaxNLocator(3, prune='both'))

    for x, ax in axes.ffill(axis=0).iloc[-1, :].dropna().iteritems():
        ax.xaxis.set_major_locator(MaxNLocator(3, prune='both'))

    return fig, axes


def plot_1d(ax, data, *args, **kwargs):
    """Plot a 1d marginalised distribution.

    This functions as a wrapper around matplotlib.axes.Axes.plot, with a kernel
    density estimation computation in between. All remaining keyword arguments
    are passed onwards.

    Parameters
    ----------
    ax: matplotlib.axes.Axes
        axis object to plot on

    data: numpy.array
        Uniformly weighted samples to generate kernel density estimator.

    xmin, xmax: float
        lower/upper prior bound
        optional, default None

    Returns
    -------
    lines: matplotlib.lines.Line2D
        A list of line objects representing the plotted data (same as
        matplotlib matplotlib.axes.Axes.plot command)

    """
    if data.max()-data.min() <= 0:
        return

    xmin = kwargs.pop('xmin', None)
    xmax = kwargs.pop('xmax', None)

    x, p = kde_1d(data, xmin, xmax)
    p /= p.max()
    i = (p >= 1e-2)

    ans = ax.plot(x[i], p[i], *args, **kwargs)
    ax.set_xlim(*check_bounds(x[i], xmin, xmax), auto=True)
    return ans


def contour_plot_2d(ax, data_x, data_y, *args, **kwargs):
    """Plot a 2d marginalised distribution as contours.

    This functions as a wrapper around matplotlib.axes.Axes.contour, and
    matplotlib.axes.Axes.contourf with a kernel density estimation computation
    in between. All remaining keyword arguments are passed onwards to both
    functions.

    Parameters
    ----------
    ax: matplotlib.axes.Axes
        axis object to plot on

    data_x, data_y: numpy.array
        x and y coordinates of uniformly weighted samples to generate kernel
        density estimator.

    xmin, xmax, ymin, ymax: float
        lower/upper prior bounds in x/y coordinates
        optional, default None

    Returns
    -------
    c: matplotlib.contour.QuadContourSet
        A set of contourlines or filled regions

    """
    xmin = kwargs.pop('xmin', None)
    xmax = kwargs.pop('xmax', None)
    ymin = kwargs.pop('ymin', None)
    ymax = kwargs.pop('ymax', None)
    color = kwargs.pop('color', next(ax._get_lines.prop_cycler)['color'])

    x, y, pdf = kde_2d(data_x, data_y,
                       xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax)
    pdf /= pdf.max()
    p = sorted(pdf.flatten())
    m = numpy.cumsum(p)
    m /= m[-1]
    interp = interp1d([0]+list(m)+[1], [0]+list(p)+[1])
    contours = list(interp([0.05, 0.33]))+[1]

    # Correct non-zero edges
    if min(p) != 0:
        contours = [min(p)] + contours

    # Correct level sets
    for i in range(1, len(contours)):
        if contours[i-1] == contours[i]:
            for j in range(i):
                contours[j] = contours[j] - 1e-5

    i = (pdf >= 1e-2).any(axis=0)
    j = (pdf >= 1e-2).any(axis=1)

    cmap = basic_cmap(color)
    zorder = max([child.zorder for child in ax.get_children()])

    cbar = ax.contourf(x[i], y[j], pdf[numpy.ix_(j, i)], contours,
                       vmin=0, vmax=1.0, cmap=cmap, zorder=zorder+1,
                       *args, **kwargs)
    for c in cbar.collections:
        c.set_cmap(cmap)

    ax.contour(x[i], y[j], pdf[numpy.ix_(j, i)], contours,
               vmin=0, vmax=1.2, linewidths=0.5, colors='k', zorder=zorder+2,
               *args, **kwargs)
    ax.set_xlim(*check_bounds(x[i], xmin, xmax), auto=True)
    ax.set_ylim(*check_bounds(y[j], ymin, ymax), auto=True)
    return cbar


def scatter_plot_2d(ax, data_x, data_y, *args, **kwargs):
    """Plot samples from a 2d marginalised distribution.

    This functions as a wrapper around matplotlib.axes.Axes.plot, enforcing any
    prior bounds. All remaining keyword arguments are passed onwards.

    Parameters
    ----------
    ax: matplotlib.axes.Axes
        axis object to plot on

    data_x, data_y: numpy.array
        x and y coordinates of uniformly weighted samples to generate kernel
        density estimator.

    xmin, xmax, ymin, ymax: float
        lower/upper prior bounds in x/y coordinates
        optional, default None

    Returns
    -------
    lines: matplotlib.lines.Line2D
        A list of line objects representing the plotted data (same as
        matplotlib matplotlib.axes.Axes.plot command)

    """
    xmin = kwargs.pop('xmin', None)
    xmax = kwargs.pop('xmax', None)
    ymin = kwargs.pop('ymin', None)
    ymax = kwargs.pop('ymax', None)

    points = ax.plot(data_x, data_y, 'o', markersize=1, *args, **kwargs)
    ax.set_xlim(*check_bounds(data_x, xmin, xmax), auto=True)
    ax.set_ylim(*check_bounds(data_y, ymin, ymax), auto=True)
    return points


def get_legend_proxy(fig):
    """Extract a proxy for plotting onto a legend.

    Example usage:
        >>> fig, axes = modelA.plot_2d()
        >>> modelB.plot_2d(axes)
        >>> proxy = get_legend_proxy(fig)
        >>> fig.legend(proxy, ['A', 'B']

    Parameters
    ----------
        fig: matplotlib.figure.Figure
            Figure to extract colors from.

    """
    cmaps = [coll.get_cmap() for ax in fig.axes for coll in ax.collections
             if isinstance(coll, PathCollection)]
    cmaps = unique(cmaps)

    if not cmaps:
        colors = [line.get_color() for ax in fig.axes for line in ax.lines]
        colors = unique(colors)
        cmaps = [basic_cmap(color) for color in colors]

    proxy = [plt.Rectangle((0, 0), 1, 1, facecolor=cmap(0.999),
                           edgecolor=cmap(0.33), linewidth=2)
             for cmap in cmaps]

    return proxy


def basic_cmap(color):
    """Construct basic colormap a single color."""
    return LinearSegmentedColormap.from_list(color, ['#ffffff', color])



"""Tools for reading from multinest chains files."""
import numpy
from anesthetic.read.getdist import GetDistReader


class MultiNestReader(GetDistReader):
    """Read multinest files."""

    def samples(self):
        """Read <root>ev.dat and <root>phys_live.points in multinest format."""
        try:
            data = numpy.loadtxt(self.birth_file)
            samples, logL, logL_birth, _ = numpy.split(data, [-4, -3, -2],
                                                       axis=1)
            logL = numpy.squeeze(logL)
            logL_birth = numpy.squeeze(logL_birth)

            data = numpy.loadtxt(self.phys_live_birth_file)
            (live_samples, live_logL,
             live_logL_birth, _) = numpy.split(data, [-3, -2, -1], axis=1)
            live_logL = numpy.squeeze(live_logL)
            live_logL_birth = numpy.squeeze(live_logL_birth)
            i = numpy.argsort(live_logL)
            samples = numpy.concatenate((samples, live_samples[i]), axis=0)
            logL = numpy.concatenate((logL, live_logL[i]))
            logL_birth = numpy.concatenate((logL_birth, live_logL_birth[i]))
            return samples, logL, logL_birth

        except IOError:
            data = numpy.loadtxt(self.ev_file)
            samples, logL, _ = numpy.split(data, [-3, -2], axis=1)
            logL = numpy.squeeze(logL)

            data = numpy.loadtxt(self.phys_live_file)
            live_samples, live_logL, _ = numpy.split(data, [-2, -1], axis=1)
            live_logL = numpy.squeeze(live_logL)
            i = numpy.argsort(live_logL)
            nlive = len(live_logL)
            samples = numpy.concatenate((samples, live_samples[i]), axis=0)
            logL = numpy.concatenate((logL, live_logL[i]))
            return samples, logL, nlive

    @property
    def ev_file(self):
        """File containing discarded points."""
        return self.root + 'ev.dat'

    @property
    def birth_file(self):
        """File containing discarded points."""
        return self.root + 'dead-birth.txt'

    @property
    def phys_live_file(self):
        """File containing physical live points."""
        return self.root + 'phys_live.points'

    @property
    def phys_live_birth_file(self):
        """File containing physical live points."""
        return self.root + 'phys_live-birth.txt'



"""Drop-in for earlier versions of matplotlib."""
from matplotlib.widgets import AxesWidget
import six
import numpy as np


class Slider(AxesWidget):
    """A slider representing a floating point range.

    Create a slider from *valmin* to *valmax* in axes *ax*. For the slider to
    remain responsive you must maintain a reference to it. Call
    :meth:`on_changed` to connect to the slider event.

    Attributes
    ----------
    val : float
        Slider value.

    """

    def __init__(self, ax, label, valmin, valmax, valinit=0.5, valfmt='%1.2f',
                 closedmin=True, closedmax=True, slidermin=None,
                 slidermax=None, dragging=True, valstep=None,
                 orientation='horizontal', **kwargs):
        """Initialisation.

        Parameters
        ----------
        ax : Axes
            The Axes to put the slider in.

        label : str
            Slider label.

        valmin : float
            The minimum value of the slider.

        valmax : float
            The maximum value of the slider.

        valinit : float, optional, default: 0.5
            The slider initial position.

        valfmt : str, optional, default: "%1.2f"
            Used to format the slider value, fprint format string.

        closedmin : bool, optional, default: True
            Indicate whether the slider interval is closed on the bottom.

        closedmax : bool, optional, default: True
            Indicate whether the slider interval is closed on the top.

        slidermin : Slider, optional, default: None
            Do not allow the current slider to have a value less than
            the value of the Slider `slidermin`.

        slidermax : Slider, optional, default: None
            Do not allow the current slider to have a value greater than
            the value of the Slider `slidermax`.

        dragging : bool, optional, default: True
            If True the slider can be dragged by the mouse.

        valstep : float, optional, default: None
            If given, the slider will snap to multiples of `valstep`.

        orientation : str, optional, default: 'horizontal'
            'horizontal' : horizontal slider
            'vertical'   : vertical slider

        Notes
        -----
        Additional kwargs are passed on to ``self.poly`` which is the
        :class:`~matplotlib.patches.Rectangle` that draws the slider
        knob.  See the :class:`~matplotlib.patches.Rectangle` documentation for
        valid property names (e.g., `facecolor`, `edgecolor`, `alpha`).

        """
        AxesWidget.__init__(self, ax)

        if slidermin is not None and not hasattr(slidermin, 'val'):
            raise ValueError("Argument slidermin ({}) has no 'val'"
                             .format(type(slidermin)))
        if slidermax is not None and not hasattr(slidermax, 'val'):
            raise ValueError("Argument slidermax ({}) has no 'val'"
                             .format(type(slidermax)))
        if orientation not in ['horizontal', 'vertical']:
            raise ValueError("Argument orientation ({}) must be either"
                             "'horizontal' or 'vertical'".format(orientation))

        self.orientation = orientation
        self.closedmin = closedmin
        self.closedmax = closedmax
        self.slidermin = slidermin
        self.slidermax = slidermax
        self.drag_active = False
        self.valmin = valmin
        self.valmax = valmax
        self.valstep = valstep
        valinit = self._value_in_bounds(valinit)
        if valinit is None:
            valinit = valmin
        self.val = valinit
        self.valinit = valinit
        if orientation == 'vertical':
            self.poly = ax.axhspan(valmin, valinit, 0, 1, **kwargs)
            self.hline = ax.axhline(valinit, 0, 1, color='r', lw=1)
        else:
            self.poly = ax.axvspan(valmin, valinit, 0, 1, **kwargs)
            self.vline = ax.axvline(valinit, 0, 1, color='r', lw=1)

        self.valfmt = valfmt
        ax.set_yticks([])
        if orientation == 'vertical':
            ax.set_ylim((valmin, valmax))
        else:
            ax.set_xlim((valmin, valmax))
        ax.set_xticks([])
        ax.set_navigate(False)

        self.connect_event('button_press_event', self._update)
        self.connect_event('button_release_event', self._update)
        if dragging:
            self.connect_event('motion_notify_event', self._update)
        if orientation == 'vertical':
            self.label = ax.text(0.5, 1.02, label, transform=ax.transAxes,
                                 verticalalignment='bottom',
                                 horizontalalignment='center')

            self.valtext = ax.text(0.5, -0.02, valfmt % valinit,
                                   transform=ax.transAxes,
                                   verticalalignment='top',
                                   horizontalalignment='center')
        else:
            self.label = ax.text(-0.02, 0.5, label, transform=ax.transAxes,
                                 verticalalignment='center',
                                 horizontalalignment='right')

            self.valtext = ax.text(1.02, 0.5, valfmt % valinit,
                                   transform=ax.transAxes,
                                   verticalalignment='center',
                                   horizontalalignment='left')

        self.cnt = 0
        self.observers = {}

        self.set_val(valinit)

    def _value_in_bounds(self, val):
        """Make sure self.val is with given bounds."""
        if self.valstep:
            val = np.round((val - self.valmin)/self.valstep)*self.valstep
            val += self.valmin

        if val <= self.valmin:
            if not self.closedmin:
                return
            val = self.valmin
        elif val >= self.valmax:
            if not self.closedmax:
                return
            val = self.valmax

        if self.slidermin is not None and val <= self.slidermin.val:
            if not self.closedmin:
                return
            val = self.slidermin.val

        if self.slidermax is not None and val >= self.slidermax.val:
            if not self.closedmax:
                return
            val = self.slidermax.val
        return val

    def _update(self, event):
        """Update the slider position."""
        if self.ignore(event):
            return

        if event.button != 1:
            return

        if event.name == 'button_press_event' and event.inaxes == self.ax:
            self.drag_active = True
            event.canvas.grab_mouse(self.ax)

        if not self.drag_active:
            return

        elif ((event.name == 'button_release_event') or
              (event.name == 'button_press_event' and
               event.inaxes != self.ax)):
            self.drag_active = False
            event.canvas.release_mouse(self.ax)
            return
        if self.orientation == 'vertical':
            val = self._value_in_bounds(event.ydata)
        else:
            val = self._value_in_bounds(event.xdata)
        if (val is not None) and (val != self.val):
            self.set_val(val)

    def set_val(self, val):
        """Set slider value to *val*.

        Parameters
        ----------
        val : float

        """
        xy = self.poly.xy
        if self.orientation == 'vertical':
            xy[1] = 0, val
            xy[2] = 1, val
        else:
            xy[2] = val, 1
            xy[3] = val, 0
        self.poly.xy = xy
        self.valtext.set_text(self.valfmt % val)
        if self.drawon:
            self.ax.figure.canvas.draw_idle()
        self.val = val
        if not self.eventson:
            return
        for cid, func in six.iteritems(self.observers):
            func(val)

    def on_changed(self, func):
        """On changed.

        When the slider value is changed call *func* with the new
        slider value

        Parameters
        ----------
        func : callable
            Function to call when slider is changed.
            The function must accept a single float as its arguments.

        Returns
        -------
        cid : int
            Connection id (which can be used to disconnect *func*)

        """
        cid = self.cnt
        self.observers[cid] = func
        self.cnt += 1
        return cid

    def disconnect(self, cid):
        """Remove the observer with connection id *cid*.

        Parameters
        ----------
        cid : int
            Connection id of the observer to be removed

        """
        try:
            del self.observers[cid]
        except KeyError:
            pass

    def reset(self):
        """Reset the slider to the initial value."""
        if (self.val != self.valinit):
            self.set_val(self.valinit)



"""Tools for reading from chains files."""


class ChainReader(object):
    """Base class for reading chains files."""

    def __init__(self, root):
        self.root = root

    def paramnames(self):
        """Parameter names mapping."""
        return None, {}

    def limits(self):
        """Parameter limits mapping."""
        return {}

    def samples(self):
        """Read samples from file."""
        raise NotImplementedError
